import mnkytw

# Declare a basic literal matcher
Hello = mnkytw.LiteralMatch("Hello ")

class MultipleHellosMatch:
    # In the class initialised
    def __init__(self):
        # declare your compound matcher
        self.matcher = mnkytw.MatchAlternation([
            mnkytw.MatchJoin([
                Hello,
                # With recursion
                self,
            ]),
            Hello,
        ])

    # Declare your parser function, and return a result
    def parser(self, body : str, hard_fail = True):
        result = self.matcher.parser(body, hard_fail)
        return result

    def to_string(self, call_count = 0):
        return self.matcher.to_string(call_count)
    
    def __str__(self):
        return self.to_string()

MultipleHellos = MultipleHellosMatch()

#parse the string "Hello Hello Hello"
print(
    mnkytw.peg_parse("Hello Hello Hello ", MultipleHellos)
)

#parse the single string "Hello "
print(
    mnkytw.peg_parse("Hello ", MultipleHellos)
)

#print the grammar as a PEG.js style grammar (ish)
print(mnkytw.to_string(MultipleHellos))



import mnkytw
class FloatMatch:
    def __init__(self):
        # A float is at least one digit before a decimal point
        # followed by any number of zero inclusive digits, and
        # finally a digit between 1 and 9
        self.matcher = mnkytw.RegexMatch(r"[0-9]+\.[0-9]*[1-9]")

    def parser(self, body : str, hard_fail = True):
        #Attempt to match to the matcher
        result = self.matcher.parser(body, hard_fail)
        # if it's not a match, and we're here, we can assume that
        # we haven't been told to "hard fail", so just return the false
        # result
        if not result:
            return result
        
        # Return your own token result, but the second value must always
        # be the cursor position within the body, so make sure you always return
        # result[1] as the second argument
        return [{
            'val': float(result[0]),
            'type': 'float'
        }, result[1]]

    def to_string(self, call_count = 0):
        return self.matcher.to_string(call_count)

    def __str__(self):
        return self.to_string()


# LiteralMatch does what it says on the tin, takes a literal such as "hello" and 
# matches it an only it. This is one of two axiomatic building blocks for a PEG parser.
# The other is RegexMatch
class LiteralMatch:
    
    # @val The value of to match to, can be any arbitrary string
    # Note that this function is case sensitive
    def __init__(self, val : str):
        self.expression = val

    # @body The body we are matching tokens to in parsing
    # @hard_fail A boolean flag indicating if a matcher should throw an error (or just return false)
    # This method parses a body of text for tokens
    def parser(self, body : str, hard_fail = True):
        #if the start of the body, to length of the expression to match
        # is the same as the expression to match then
        match = body[0:len(self.expression)] == self.expression
        #if not a match
        if not match:
            #If set to hard fail
            if hard_fail:
                #throw an error
                raise ValueError(f"LiteralMatch: Could not match '{self.expression}'")
            else:
                #else return false
                return False
        #if matched
        #return the expression and the length of the head as a tuple pair
        return (self.expression, len(self.expression))

    # @call_count A nonce used to detect recursion
    def to_string(self, call_count = 0):
        return f'"{self.expression}"'

    def __str__(self):
        return self.to_string()


# MatchJoin takes a list of matchers and joins them, to essentially
# make a conjunctive matcher
class MatchJoin:
    # @matchers A list of matchers to join into a single match for example
    # [
    #   LiteralMatch("hello ")
    #   LiteralMatch("world")
    # ]
    def __init__(self, matchers : list):
        self.to_string_call_count = None
        self.matchers = matchers

    # @body The body to parse and tokenise
    # @hard_fail a boolean flag indicating whether or not to throw an error when a match
    # doesn't work
    def parser(self, body : str, hard_fail = True):
        result = []
        head = 0
        for matcher in self.matchers:
            sub_body = body[head:]
            out = matcher.parser(sub_body, hard_fail = False)
            if not out:
                if hard_fail:
                    raise ValueError(f"MatchJoin: Could not match at {head}")
                else:
                    return False
            result.append(out[0])
            head += out[1]
        return (result, head)
    
    def to_string(self, call_count = 0):    
        if self.to_string_call_count == call_count:
            return "..."
        else:
            self.to_string_call_count = call_count
            return "(" + " ".join([m.to_string(call_count) for m in self.matchers]) + ")"

    def __str__(self):
        return self.to_string()



import mnkytw
class IntegerMatch:
    def __init__(self):
        #Match either
        self.matcher = mnkytw.MatchAlternation([
            #A 1 to 9 followed by any number of digits
            mnkytw.RegexMatch(r"[1-9][0-9]*"),
            #or a zero
            mnkytw.LiteralMatch("0")
        ])

    def parser(self, body : str, hard_fail = True):
        #Attempt to match to the matcher
        result = self.matcher.parser(body, hard_fail)
        # if it's not a match, and we're here, we can assume that
        # we haven't been told to "hard fail", so just return the false
        # result
        if not result:
            return result
        
        # Return your own token result, but the second value must always
        # be the cursor position within the body, so make sure you always return
        # result[1] as the second argument
        return [{
            'val': int(result[0]),
            'type': 'integer'
        }, result[1]]

    def to_string(self, call_count = 0):
        return self.matcher.to_string(call_count)

    def __str__(self):
        return self.to_string()


# MatchAlternation takes a list of matchers and tries them in order
# until one matches the given body. 
# Note that because we don't enforce strict end matching, MatchAlternation 
# will act greedily essentially using the first # match and consuming as 
# much as possible with that match
class MatchAlternation:
    # @matchers A list of matchers in the order to attempt matching
    def __init__(self, matchers : list):
        self.to_string_call_count = None
        self.matchers = matchers
    
    # @body The body we are matching tokens to in parsing
    # @hard_fail A boolean flag indicating if a matcher should throw an error (or just return false)
    # This method parses a body of text for tokens
    def parser(self, body : str, hard_fail = True):
        for matcher in self.matchers:
            result = matcher.parser(body, hard_fail = False)
            if result:
                return result
        if hard_fail:
            raise ValueError(f"MatchAlternation: Could not match")
        else: 
            return False

    # @call_count A nonce used to detect recursion
    def to_string(self, call_count = 0):
        if self.to_string_call_count == call_count:
            return "..."
        else:
            self.to_string_call_count = call_count
            return "(" + " / ".join([str(m) for m in self.matchers]) + ")"

    def __str__(self):
        return self.to_string(self)


class MatchQuantity:
    def __init__(self, matcher, min_n = 0, max_n = -1):
        self.matcher = matcher
        self.min = min_n
        self.max = max_n

    def parser(self, body : str, hard_fail = True):
        head = 0
        count = 0
        result = []
        while True:
            sub_body = body[head:]
            # If max is specified and we've reached it
            if self.max > 0 and self.max == count:
                # return the result and the head
                return (result, head)
            match = self.matcher.parser(sub_body, hard_fail = False)
            if not match or head == len(body):
                break
            head += match[1]
            result.append(match[0])
            count += 1
        
        if count < self.min:
            if hard_fail:
                raise ValueError(f"MatchQuantity: less than minimum matches")
            else:
                return False
        
        return (result, head)

    def to_string(self, call_count = 0):
        return f"({self.matcher.to_string(call_count)}){{{self.min},{self.max}}}"

    def __str__(self):
        return self.to_string()


import mnkytw
from examples.maths.IntegerMatch import IntegerMatch
from examples.maths.FloatMatch import FloatMatch

# Create a single unified matcher that attempts to identify
# an integer or a float
Constants = mnkytw.MatchAlternation([
    FloatMatch(),
    IntegerMatch()
])

# symbols
Symbols = mnkytw.MatchAlternation([
    mnkytw.LiteralMatch("+"),
    mnkytw.LiteralMatch("-"),
    mnkytw.LiteralMatch("*"),
    mnkytw.LiteralMatch("/")
])

class OperationMatch:
    def __init__(self):
        self.matcher = mnkytw.MatchAlternation([
            # Either this is a chain of operations
            mnkytw.MatchJoin([
                Constants,
                Symbols,
                self
            ]),
            # Or a constant
            Constants
        ])
    
    def parser(self, body : str, hard_fail = True):
        result = self.matcher.parser(body, hard_fail)
        if not result:
            return result
        
        #we can infer that it matched the "MatchJoin" if it's a list
        if type(result[0]) is list:
            # make a dictionary that shows the lhs and rhs and the symbol
            return [{
                'lhs': result[0][0],
                'symbol': result[0][1],
                'rhs': result[0][2]
            }, result[1]]
        else:
            #Otherwise it matched the constant so just return the constant
            return result
        
Operation = OperationMatch()

print(mnkytw.peg_parse("3+4", Operation))

print(mnkytw.peg_parse("3+4+5", Operation))

print(mnkytw.peg_parse("3+4*5-6", Operation))


import random
from mnkytw.LiteralMatch import LiteralMatch
from mnkytw.MatchAlternation import MatchAlternation
from mnkytw.MatchJoin import MatchJoin
from mnkytw.MatchQuantity import MatchQuantity
from mnkytw.RegexMatch import RegexMatch

# @body the body to parse (fully)
# @matcher the matcher to use to parse the body
# This function is a helper function designed to execute a matcher
# and test if said matcher fully matches the body
# Generally speaking the matcher should be a document matcher, capable
# of fully parsing and account for all possible ways for expressing your
# language, including appropriate whitespacing
def peg_parse(body : str, matcher):
    result = matcher.parser(body, False)
    if result[1] != len(body):
        return False
    return result[0]

def to_string(matcher):
    return matcher.to_string(random.random())


import re

class RegexMatch:
    def __init__(self, reg : str):
        #We store the regex to make printing easier
        self.expression = reg
        #compile it into an actual regex though
        self.regexr = re.compile(reg)

    def parser(self, body : str, hard_fail = True):
        #match the body, to the regex 
        # (we use match because we're essentially start anchoring)
        match = self.regexr.match(body)
        # if not a match
        if not match:
            # if hard fail is true
            if hard_fail:
                # throw an error
                raise ValueError(f"RegexMatch: Could not match '{self.expression}'")
            else:
                #otherwise just return false
                return False
        #return the match and the head of the string (i.e. where in the body we are now)
        return (match.group(0), match.span()[1])

    def to_string(self, call_count = 0):
        return f'"{self.expression}"'

    def __str__(self):
        return self.to_string()


import examples.hello
import examples.maths

import mnkytw

matchNoMoreThanFive3s = mnkytw.MatchQuantity(
    mnkytw.LiteralMatch("3"),
    0, #Can match no threes
    5 # match at most 5 threes
)

print(matchNoMoreThanFive3s.parser("33", False))
# (['3', '3'], 2)

print(matchNoMoreThanFive3s.parser("333333", False))
# (['3', '3', '3', '3', '3'], 5)

print(matchNoMoreThanFive3s.parser("343", False))
# (['3'], 1)

print(matchNoMoreThanFive3s.parser("245", False))
# ([], 0)


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='monkeys-typewriter',
      version=open('VERSION').read(),
      description="Monkey's Typewriter PEG parser",
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Adam Green',
      author_email='adam@tryal.ai',
      donwload_url='https://github.com/tryal-ai/mnkytw/archive/1.0.1.tar.gz',
      url='https://tryal.ai/',
      packages=find_packages(),
      keywords=["PEG", "parser", "grammar"],
      python_requires='>=3.6',
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)